// List all files in a NGC ISO file.
// Usage :
// ListFST <filename.gcm>

/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>
#include <sys/types.h>
#include <dirent.h>

#define BUFFERSIZE 1028

//! The mount point.
static char mount_point[] = "/tmp/XXXXXX";

//! File descriptor of the game image.
static int fd_game = -1;

/**
 * Open the game file.
 * Automaticly detect the format (iso/ziso).
 * \param filename the name of the game file.
 * \return the file descriptor of the game, else -1.
 */
int open_game(char *filename)
{
	int fd = 0;
	int retcode = 0;
	uint8_t buffer[BUFFERSIZE];
	uint8_t command[1024+1];
	DIR *dir = NULL;
	struct dirent *dirent = NULL;
	char *file_game = NULL;

	fd = open(filename, O_RDONLY);
	if(-1 == fd){
		perror(filename);
		return -1;
	}
	
	retcode = lseek(fd, 32808, SEEK_SET);
	if(-1 == retcode) {
		perror(filename);
		return -1;
	}

	retcode = read(fd, buffer, 5);
	if(-1 == retcode) {
		perror(filename);
		return -1;
	}

	//TODO to be improved
	if(!strncmp("CDROM", buffer, 5)) {
		printf("Compressed ISO9660 File System detected\n");
		if(NULL == mkdtemp(mount_point)) {
			perror(mount_point);
			return -1;
		}

		//TODO to be improved
		memset(command, 0, 1024+1);
		snprintf(command, 1024, "mount -o loop %s %s", filename, mount_point);
		retcode = system(command);
		if(retcode != 0) {
			fprintf(stderr, "unable to mount the file system\n");
			return -1;
		}
		
		dir = opendir(mount_point);
		if(NULL == dir) {
			perror(mount_point);
			return -1;
		}

		//TODO to be improved
		dirent = readdir(dir); //.
		dirent = readdir(dir); //..
		dirent = readdir(dir); //the game
		if(NULL == dirent) {
			perror(mount_point);
			return -1;
		}
		
		memset(command, 0, 1024+1);
		snprintf(command, 1024, "%s/%s", mount_point, dirent->d_name);

		retcode = closedir(dir);
		if(-1 == retcode) {
			perror(mount_point);
			return -1;
		}

		file_game = command;
	}
	else file_game = filename;

	fd = open(file_game, O_RDONLY);
	if(-1 == fd) {
		perror(filename);
		return -1;
	}

	return fd;
}

/**
 * Terminat properly the program.
 * Close all open resources.
 * \param the sigan lnumber.
 */
void handler(int i)
{
	int retcode = 0;
	char command[1024+1];

	fprintf(stderr, "\rExiting\n");

	if(strncmp(mount_point, "/tmp/XXXXXX", 11)) {
		//TODO to be improved
		snprintf(command, 1024, "umount %s", mount_point);
		retcode = system(command);
		if(retcode != 0) {
			perror(mount_point);
			return;
		}
		
		retcode = rmdir(mount_point);
		if(-1 == retcode) {
			perror(mount_point);
			return;
		}
	}

	exit(1);
}


int main(int argc, char **argv)
{
	unsigned char buffer[256];
	unsigned char *FST;
	unsigned long fst_offset=0,filename_offset=0,entries=0,string_table_offset=0,fst_size=0;
	unsigned int  found=0;
	unsigned long i=0,j=0,offset=0,parent_offset=0,next_offset=0,name_offset=0;
	unsigned char filename[30], path[266], temp[256], current_path[256];

	if(argc!=2)
	{
		printf("usage: %s <filename.[z]iso>\n",argv[0]);
		return -1;
	}

	strcpy(filename,"");
	strcpy(path,"");
	strcpy(current_path,"/");

	fd_game = open_game(argv[1]);
	if(-1 == fd_game) {
		perror(argv[1]);
		exit(1);
	}

	read(fd_game, buffer, 256);
	buffer[255]=0;
	printf("%c%c%c (",buffer[0],buffer[1],buffer[2]);
	switch(buffer[3]){
	case 'E':
		printf("USA/NTSC");
		break;
	case 'P':
		printf("Europe/PAL");
		break;
	default:   
		printf("Japan/NTSC");
	}
	printf(") - %s\n",buffer+32);

	// Load all the FST in memory :
	lseek(fd_game,0x424,SEEK_SET);
	read(fd_game,buffer,8);
	fst_offset=(unsigned int)buffer[0]*256*256*256+(unsigned int)buffer[1]*256*256+(unsigned int)buffer[2]*256+(unsigned int)buffer[3];
	//      printf("Fst offset=%ld - ",fst_offset);
	fst_size=(unsigned int)buffer[4]*256*256*256+(unsigned int)buffer[5]*256*256+(unsigned int)buffer[6]*256+(unsigned int)buffer[7];
	FST=malloc(fst_size);
	lseek(fd_game,fst_offset,SEEK_SET);
	read(fd_game,FST,fst_size);
	close(fd_game);

	entries=(unsigned int)FST[8]*256*256*256+(unsigned int)FST[9]*256*256+(unsigned int)FST[10]*256+(unsigned int)FST[11];
	string_table_offset=12*entries;
	//      printf("Number of FST entries : %i\n",entries);

	for (i=1;i<entries;i++)
	{
		offset=i*0x0c;
		if(FST[offset]==0)
		{
			filename_offset=(unsigned int)FST[offset+1]*256*256+(unsigned int)FST[offset+2]*256+(unsigned int)FST[offset+3];
			strcpy(filename,&FST[string_table_offset+filename_offset]);

			// Looking for the full path of the file :
			strcpy(path,"");
			strcpy(temp,"");
			j=0x0c;
			found=0;
			while ( (j<offset) && (found==0) )
			{
				// Does the file appear in a "next offset" reference ?
				// (search starts from the begining of the FST)
				if(FST[j]==1)
				{
					next_offset=(unsigned int)FST[j+8]*256*256*256+(unsigned int)FST[j+9]*256*256+(unsigned int)FST[j+10]*256+(unsigned int)FST[j+11];
					if (next_offset==i)
					{
						parent_offset=(unsigned int)FST[j+4]*256*256*256+(unsigned int)FST[j+5]*256*256+(unsigned int)FST[j+6]*256+(unsigned int)FST[j+7];
						while(parent_offset!=0)
						{
							// Folder's name :
							name_offset=(unsigned int)FST[parent_offset*0x0c+1]*256*256+(unsigned int)FST[parent_offset*0x0c+2]*256+(unsigned int)FST[parent_offset*0x0c+3];
							strcpy(temp,&FST[string_table_offset+name_offset]);
							strcat(temp,"/");
							strcat(temp,path);
							strcpy(path,temp);

							parent_offset=(unsigned int)FST[parent_offset*0x0c+4]*256*256*256+(unsigned int)FST[parent_offset*0x0c+5]*256*256+(unsigned int)FST[parent_offset*0x0c+6]*256+(unsigned int)FST[parent_offset*0x0c+7];
						}
						strcpy(temp,"/");
						strcat(temp,path);
						strcpy(path,temp);
						found=1;
					}
				}
				j=j+0x0c;
			}
			// If the file is not in a "next offset" reference
			//      check the if the previous entry is not a directory
			//      if YES path=path of the previous entry (which is a folder)
			//      if NO path=curent path (=the path of the previous file)
			// else
			//      path=path previously found
			if(found==1)
			{
				strcpy(current_path,path);
			}
			else
			{
				if ((offset>0x0c) && (FST[offset-0x0c]==1))
				{
					parent_offset=i-1;
					while(parent_offset!=0)
					{
						// Folder's name :
						name_offset=(unsigned int)FST[parent_offset*0x0c+1]*256*256+(unsigned int)FST[parent_offset*0x0c+2]*256+(unsigned int)FST[parent_offset*0x0c+3];
						strcpy(temp,&FST[string_table_offset+name_offset]);
						strcat(temp,"/");
						strcat(temp,path);
						strcpy(path,temp);

						parent_offset=(unsigned int)FST[parent_offset*0x0c+4]*256*256*256+(unsigned int)FST[parent_offset*0x0c+5]*256*256+(unsigned int)FST[parent_offset*0x0c+6]*256+(unsigned int)FST[parent_offset*0x0c+7];
					}
					strcpy(temp,"/");
					strcat(temp,path);
					strcpy(path,temp);
					strcpy(current_path,path);
				}
			}
			printf("%ld %s%s\n", string_table_offset+filename_offset, current_path, filename);
		}
	}

	handler(0);
	return 0;
} 
